//
//------------------------------------------------------------------------------
//  Name    : hwsuart_HD.bas
//  Author  : Jerry Messina
//  Date    : 10/26/2018
//  Version : 2.0.0
//------------------------------------------------------------------------------
//
//  Hardware-assisted software uart (hwsuart half-duplex)
//
//  This module is an interrupt-driven 'software' uart that uses hardware peripheral
//  functions to assist in the generation of the serial data timing and detection of
//  the RX start bit. In addition, it provides an input RX buffer to hold incoming data
//  allowing the module to be used for asyncronous reception (which is something most
//  software uarts cannot do).
//
//  Resources Used
//    - high-priority interrupt
//    - INTx edge-triggered interrupt input pin for RXD (selectable INT0/INT1/INT2 or IOC RB4)
//    - CCPx module and associated TMRx (compare mode, selectable CCP1/CCP2 and TMR1/3/5)
//    - TXD pin (any PIO pin)
//    - optional
//        RS485 transceiver direction control (any PIO pin)
//        DEBUG trigger output (any PIO pin)
//
//  The module uses one of the INTx pins for RXD. It is set to provide an interrupt on
//  the falling edge of RXD to detect the initial receive START bit, and is then sampled
//  at an interval set by the CCPx module to assemble the serial input data. When a
//  START bit is detected, the CCP module is setup for 1/2 of the serial bit time,
//  and will then generate an interrupt to sample the incoming RX data stream approx
//  in the middle of the bit time.
//
//  The RX byte is assembled and then placed into a receive buffer, whose size can
//  be set via the '#option HWSUART_RX_BUFFER_SIZE' to be from 4 to 128 bytes (in powers
//  of 2). Functions are provided to check the status and access the RX buffer, including
//  data_available(), buffer_count(), flush_input(), and read_byte().
//
//  TX data output is also controlled by the CCP module timing. Unlike the RX data,
//  the TX data out can reside on any port pin. There is no output queue for sending
//  data... you must call write_byte() for each byte you wish to transmit.
//
//  Serial data format is N81, with baud rates from 1200 to 115200 depending on the
//  clock speed. Supported mpu clock rates are 8, 10, 16, 20, 32, 40, and 64MHz.
//  Slower clock rates (8/10MHz) support up to 38.4K baud, intermediate rates
//  (16/20MHz) up to 57.6K, and faster rates (32/40/64MHz) up to 115.2K baud
//
//  The hwsuart operates in half-duplex mode only (you would need a separate rx and
//  tx bit timer to support full-duplex operation). Be aware that since it operates
//  in half-duplex mode, it's possible to drop incoming characters or even get a false
//  corrupt byte (a partial char) if the module is busy transmitting when a receive
//  char comes in.
//
//------------------------------------------------------------------------------
//
module hwsuart

// input buffer size (must be a power of 2 between 4 and 128)
#option HWSUART_RX_BUFFER_SIZE = 8      // set default buffer size
#if not (HWSUART_RX_BUFFER_SIZE in (4, 8, 16, 32, 64, 128))
  #error HWSUART_RX_BUFFER_SIZE, "invalid buffer size option"
#endif
const RX_BUFFER_SIZE = HWSUART_RX_BUFFER_SIZE

//
//------------------------------------------------------------------------
// hardware pins
// change these to match your hardware
//------------------------------------------------------------------------
//
// INTx, TMRx, and CCPx default selections
#option HWSUART_INT_SELECT = 1      // RXD pin 0=INT0, 1=INT1, 2=INT2, 4=IOCB4 (RB4)
#option HWSUART_TMR_SELECT = 1      // TMR 1/3/5
#option HWSUART_CCP_SELECT = 1      // CCP 1/2

// for devices without PPS, the INT_SELECT option will determine the RXD pin
// otherwise you can use PPS to map it and select the pin via #option HWSUART_RXD_PIN
#if (HWSUART_INT_SELECT = 0)
  #option HWSUART_RXD_PIN = PORTB.0    // INT0 pin
#elseif (HWSUART_INT_SELECT = 1)
  #option HWSUART_RXD_PIN = PORTB.1    // INT1 pin
#elseif (HWSUART_INT_SELECT = 2)
  #option HWSUART_RXD_PIN = PORTB.2    // INT2 pin
#elseif (HWSUART_INT_SELECT = 4)
  #option HWSUART_RXD_PIN = PORTB.4    // IOCB4 pin
#else
  #error HWSUART_INT_SELECT, "invalid RXD/INTx pin"
#endif

// software uart rxd in (default if not set by user/INT_SELECT)
#option HWSUART_RXD_PIN = PORTC.7
#if isvalidportpin(HWSUART_RXD_PIN)
  dim RXD_PIN as HWSUART_RXD_PIN.HWSUART_RXD_PIN@
#else
  #error HWSUART_RXD_PIN, "invalid port pin option"
#endif

// software uart txd out
#option HWSUART_TXD_PIN = PORTC.6
#if isvalidportpin(HWSUART_TXD_PIN)
  dim TXD_PIN as HWSUART_TXD_PIN.HWSUART_TXD_PIN@
#else
  #error HWSUART_TXD_PIN, "invalid port pin option"
#endif

// optional pins
// RS485 direction control (2-wire half-duplex SN75176 style DE pin)
// sn75176 pins: 2=RE_N receive enable (active low), 3=DE transmit enable (active high)
// usage example: #option HWSUART_RS485_TXE = PORTC.5
#if isoption(HWSUART_RS485_TXE)
  #if isvalidportpin(HWSUART_RS485_TXE)
    dim RS485_TXE as HWSUART_RS485_TXE.HWSUART_RS485_TXE@
  #else
    #error HWSUART_RS485_TXE, "invalid port pin option"
  #endif
#endif  // isoption(HWSUART_RS485_TXE)

// enable debug pin output (optional)
// the debug pin can be used to measure the rx bit timing. it is pulsed when
// the rx startbit is detect and then asserted high when the rx pin is sampled
// and remain high for the duration of the rx routine. 
// to measure the rx timing, send alternating bit pattern (ie 'U' $55) and look
// at the DEBUG pin vs the RXD pin on a scope. ideally the RXD pin will be sampled
// at 50%-75% of the bit time.
#if isoption(HWSUART_DEBUG_PIN)
  #if isvalidportpin(HWSUART_DEBUG_PIN)
    #warning "enabling HWSUART_DEBUG_PIN output"
    dim DEBUG_PIN as HWSUART_DEBUG_PIN.HWSUART_DEBUG_PIN@    // debug output (for scope)
  #else
    #error HWSUART_DEBUG_PIN, "invalid port pin option"
  #endif
#endif  // isoption(HWSUART_DEBUG_PIN)

// TX pacing default, in usecs
// here you can add additional time after the tx stop bit is sent in case the receiving
// device needs some time in between bytes. can be changed at runtime by setting 'tx_pacing'
#option HWSUART_TX_PACING = 0
const DEFAULT_TX_PACING as word = HWSUART_TX_PACING

// interrupt priority level
// since the bit timing is critical, only the high-priority intr should be used
#option HWSUART_PRIORITY_LEVEL = 2      // 1=low, 2=high
#if not(HWSUART_PRIORITY_LEVEL in (1, 2))
  #error HWSUART_PRIORITY_LEVEL, "intr priority must be 1 or 2"
#endif  
const PRIORITY_LEVEL = HWSUART_PRIORITY_LEVEL

//
//------------------------------------------------------------------------
// INTx, TMRx, and CCPx register bit definitions
//------------------------------------------------------------------------
//
// the following register/bit definitions are device dependent
// definitions shown are for the 2xK22/4xK22 family
// if you change devices then you may have to modify the routines in the
// 'hardware setup/control functions' section as well
//
#if not(_device in(18F23K22, 18F24K22, 18F25K22, 18F26K22) or _
        _device in(18F43K22, 18F44K22, 18F45K22, 18F46K22))
  #warning "INT, TMR, and CCP register definitions may need modification for " + _device
#endif

//
// select INTx (0, 1, 2, 4)
//
#option HWSUART_INT_SELECT = 0
#if not(HWSUART_INT_SELECT in (0, 1, 2, 4))
  #error HWSUART_INT_SELECT, "unsupported INT select"
#endif
const INT_NO = HWSUART_INT_SELECT

#if (HWSUART_INT_SELECT = 0)
dim
    INTxIF  as INTCON.bits(1),      // INT0IF
    INTxIE  as INTCON.bits(4),      // INT0IE
    INTxIP  as WREG.bits(0),        // dummy bit (INT0 is always high priority)
    INTxEDG as INTCON2.bits(6)      // INTEDG0

#elseif (HWSUART_INT_SELECT = 1)
dim
    INTxIF  as INTCON3.bits(0),     // INT1IF
    INTxIE  as INTCON3.bits(3),     // INT1IE
    INTxIP  as INTCON3.bits(6),     // INT1IP
    INTxEDG as INTCON2.bits(5)      // INTEDG1

#elseif (HWSUART_INT_SELECT = 2)
dim
    INTxIF  as INTCON3.bits(1),     // INT2IF
    INTxIE  as INTCON3.bits(4),     // INT2IE
    INTxIP  as INTCON3.bits(7),     // INT2IP
    INTxEDG as INTCON2.bits(4)      // INTEDG2

#elseif (HWSUART_INT_SELECT = 4)
dim
    INTxIF  as INTCON.bits(0),      // RBIF
    INTxIE  as INTCON.bits(3),      // RBIE
    INTxIP  as INTCON2.bits(0),     // RBIP
    INTxEDG as PORTB.bits(4)        // not really the edge in this mode... it's the IOC pin
#endif  // HWSUART_INT_SELECT

//
// select TMR (1/3/5)
//
#option HWSUART_TMR_SELECT = 3
#if not(HWSUART_TMR_SELECT in (1, 3, 5))
  #error HWSUART_TMR_SELECT, "unsupported TMR select"
#endif
const TMR_NO = HWSUART_TMR_SELECT

#if (HWSUART_TMR_SELECT = 1)
// TMR1
dim TxCON as T1CON,
    TMRxH as TMR1H,
    TMRxL as TMR1L

dim TMRxIF  as PIR1.bits(0),
    TMRxIE  as PIE1.bits(0),
    TMRxIP  as IPR1.bits(0)

#elseif (HWSUART_TMR_SELECT = 3)
// TMR3
dim TxCON as T3CON,
    TMRxH as TMR3H,
    TMRxL as TMR3L
    
dim TMRxIF as PIR2.bits(1),
    TMRxIE as PIE2.bits(1),
    TMRxIP as IPR2.bits(1)

#elseif (HWSUART_TMR_SELECT = 5)
// TMR5
dim TxCON as T5CON,
    TMRxH as TMR5H,
    TMRxL as TMR5L
    
dim TMRxIF as PIR5.bits(1),
    TMRxIE as PIE5.bits(1),
    TMRxIP as IPR5.bits(1)
#endif  // HWSUART_TMR_SELECT

// TxCON register bits
// **bits in this register are device dependent**
dim
    TMRxCS1 as TxCON.bits(7),
    TMRxCS0 as TxCON.bits(6),
    TxCKPS1 as TxCON.bits(5),
    TxCKPS0 as TxCON.bits(4),
    TxOSCEN as TxCON.bits(3),
    TxSYNC  as TxCON.bits(2),
    TxRD16  as TxCON.bits(1),
    TMRxON  as TxCON.bits(0)

// TxCON setting (K22)
//  TMRxCS1:TMRxCS0     00= Timerx clock source is the instruction clock (FOSC/4)
//  TxCKPS1:TxCKPS0     00= 1:1 Prescale value
//  TxOSCEN              0= SOSC/SCLKI are disabled for Timerx and TxCKI is enabled
//  TxSYNC               0= This bit is ignored; Timer uses the internal clock
//  TxRD16               1= Enables register read/write of Timerx in one 16-bit operation
//  TMRxON               0= Stops Timerx
const
    TMR_16BIT_INT_PRESCALER_OFF = %00000010

// write 16-bit timer high-byte/low-byte
macro write_tmr16(wval)
    TMRxH = (wval >> 8)
    TMRxL = byte(wval)
end macro

macro read_tmr16(wresult)
    wresult.byte0 = TMRxL
    wresult.byte1 = TMRxH
end macro

dim TMRx as TMRxL.asword

//
// select CCP (1/2)
//
#option HWSUART_CCP_SELECT = 2
#if not(HWSUART_CCP_SELECT in (1, 2))
  #error HWSUART_CCP_SELECT, "unsupported CCP select"
#endif
const CCP_NO = HWSUART_CCP_SELECT

#if (HWSUART_CCP_SELECT = 1)
// CCP1
dim CCPxCON as CCP1CON,
    CCPRxL  as CCPR1L,
    CCPRxH  as CCPR1H,
    CCPx    as CCPR1L.asword

dim CCPxIF  as PIR1.bits(2),
    CCPxIE  as PIE1.bits(2),
    CCPxIP  as IPR1.bits(2)

#elseif (HWSUART_CCP_SELECT = 2)
// CCP2
dim CCPxCON as CCP2CON,
    CCPRxL  as CCPR2L,
    CCPRxH  as CCPR2H,
    CCPx    as CCPR2L.asword

dim CCPxIF  as PIR2.bits(0),
    CCPxIE  as PIE2.bits(0),
    CCPxIP  as IPR2.bits(0)
#endif      // HWSUART_CCP_SELECT

// CCPxCON register bits
dim
    DCxB1   as CCPxCON.bits(5),
    DCxB0   as CCPxCON.bits(4),
    CCPxM3  as CCPxCON.bits(3),
    CCPxM2  as CCPxCON.bits(2),
    CCPxM1  as CCPxCON.bits(1),
    CCPxM0  as CCPxCON.bits(0)

// CCPxCON setting (K22)
//  bits 7:6            00= unused
//  DCxB1:DCxB0         00= unused in compare mode
//  CCPxM3:CCPxM0     1011= compare mode, special event trigger, timer reset
const
    CCP_COMPARE_SPEC_EVENT_TRIG = %00001011

//
//------------------------------------------------------------------------
// uart state/status register
//------------------------------------------------------------------------
//
public structure suart_status_reg_t
    b as byte
    wait_for_start  as b.bits(0)        // edge detect enabled (idle, waiting)
    rx              as b.bits(1)        // receiving (busy)
    tx              as b.bits(2)        // tx request (busy)
    tx_send_stop    as b.bits(3)        // tx sending stop bit
    tx_done         as b.bits(4)        // tx complete
    spare1          as b.bits(5)        // unused
    frame_err       as b.booleans(6)    // invalid start/stop bit
    overrun         as b.booleans(7)    // too many rcv chars in rxbuf. input blocks
end structure

public dim status_reg as suart_status_reg_t

//
//------------------------------------------------------------------------
// tx/rx variables and buffers
//------------------------------------------------------------------------
//
// the receive data buffer is implemented as a simple circular queue.
// data is placed into the queue at 'rxhead', and is read from the queue at 'rxtail'.
// in this simple implementation, no additional status tracking flags are used (ie
// no buffer full/empty), and as a result you lose one position in the buffer which
// is used to signify that the buffer is empty (when rxhead = rxtail). The size of the
// buffer can be changed with the '#option HWSUART_RX_BUFFER_SIZE', which defaults to 8
// if not specified, so the buffer can hold up to 7 bytes of data by default.
//
public const HWSUART_MAX_CHARS = RX_BUFFER_SIZE-1   // max number of chars in the buffer
const RX_BUFFER_MASK = RX_BUFFER_SIZE-1             // buffer size MUST be a power of 2

dim
    rxd as byte,                    // rx in shift reg
    txd as byte,                    // tx out shift reg
    bit_count as byte,              // current bit number (start + 8 data + stop)
    txd_out as bit                  // tx IO pin output data

// baud rate timer settings
structure timer_data_t
    baudrate            as longword // current baudrate
    bit_time            as word     // full bit time in tmr ticks
    half_bit_time       as word     // initial half-bit time in tmr ticks
end structure

dim
    tmr_data as timer_data_t        // tmr0 count values (from table)

// additional time after the tx stop bit is sent
public dim
    tx_pacing as word               // delay, in usecs
    
// rx buffer
dim
    rxb as byte,                            // rx byte to add to buffer
    rxhead as byte,                         // buf indexes... data goes in at the head
    rxtail as byte,                         // and is read out the tail
    rxbuf(RX_BUFFER_SIZE) as byte           // receive buffer

public dim
   rx_term as char                          // read string terminator

//
//------------------------------------------------------------------------
// uart baud rate bit times (in usecs)
//------------------------------------------------------------------------
//
// NOTE:
// at data rates of 38400 and above, there may not be sufficient processing time
// to handle back-to-back reception using 1 stop bit. if reception is garbled, try
// setting the transmitter's uart to 2 stop bits, or add a small delay between bytes.
//
// uart bit times, in usecs (for ref)
//   1200 = 833.33us/bit
//   2400 = 416.66us
//   4800 = 208.33us
//   9600 = 104.16us
//  19200 =  52.08us
//  28800 =  34.72us
//  38400 =  26.04us
//  57600 =  17.36us
// 115200 =   8.68us
// 230400 =   4.34us

// clock dependent baud rate timings
const CPU_CLK = _clock                      // system clock freq
const TMR_CLK = (CPU_CLK * 1000000.0)/4     // timer clock, in MHz (FOSC/4)

// Start Bit OverHead instruction cycles
// number of instructions from start bit detected intr to sampling of start bit
// TIME: rx start bit falling edge to TMRxON = ~15 instruction cycles
// TIME: from start bit TMRxON to data sample= ~13 instruction cycles (with half_bit_time=1)
const SBOH = 28

// overhead cycles
const OH = 1

// uart baud rate bit times, in tmr clocks (ie instruction cycles)
const
   BRG_1200     as word = TMR_CLK / 1200 - OH,      // 1200 baud
   BRG_2400     as word = TMR_CLK / 2400 - OH,      // 2400 baud
   BRG_4800     as word = TMR_CLK / 4800 - OH,      // 4800 baud
   BRG_9600     as word = TMR_CLK / 9600 - OH,      // 9600 baud
   BRG_19200    as word = TMR_CLK / 19200 - OH,     // 19200 baud
   BRG_38400    as word = TMR_CLK / 38400 - OH,     // 38400 baud
   BRG_57600    as word = TMR_CLK / 57600 - OH,     // 57600 baud (min clock 16MHz)
   BRG_115200   as word = TMR_CLK /115200 - OH,     // 115200 baud (min clock 32MHz)
   BRG_230400   as word = TMR_CLK /230400 - OH      // 230400 baud (min clock 64MHz)

// RX half-bit times (used to find the middle of the START bit)
// normally these would be 1/2 of the BRG_XXXX bit time, but some combinations
// of slower clock speeds and faster baud rates require adjustments.
// setting the half_bit_time = 1 will result in the start bit timer generating
// a CCPIF one instruction cycle later while still inside the isr.
//
// these are supported at all clock frequencies with no special considerations
const
   HBT_1200  as word = BRG_1200/2,
   HBT_2400  as word = BRG_2400/2,
   HBT_4800  as word = BRG_4800/2,
   HBT_9600  as word = BRG_9600/2,
   HBT_19200 as word = BRG_19200/2,
   HBT_38400 as word = BRG_38400/2

// clock dependent timings
#if (_clock = 8)
const
   HBT_57600 as word = 0,               // 16, not supported
   HBT_115200 as word = 0               // 8, not supported

#elseif (_clock = 10)
const
   HBT_57600 as word = 1,               // 21, borderline skip rx half_bit_time intr
   HBT_115200 as word = 0               // 10, not supported

#elseif (_clock = 16)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = 0               // 16, not supported

#elseif (_clock = 20)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = 1               // 21, borderline skip rx half_bit_time intr

#elseif (_clock = 32)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = BRG_115200/2

#elseif (_clock = 40)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = BRG_115200/2

#elseif (_clock = 64)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = BRG_115200/2
#else
  #error "unsupported clock freq"
#endif

//
//------------------------------------------------------------------------
// debug pin support
//------------------------------------------------------------------------
//
inline sub HWSUART_DEBUG_PIN_ON()
  #if isoption(HWSUART_DEBUG_PIN)
    DEBUG_PIN = 1
  #endif
end sub

inline sub HWSUART_DEBUG_PIN_OFF()
  #if isoption(HWSUART_DEBUG_PIN)
    DEBUG_PIN = 0
  #endif
end sub

//
//------------------------------------------------------------------------
// RS485 mode support (optional)
//------------------------------------------------------------------------
//
macro RS485_TX_ENABLE(x)
  #if isoption(HWSUART_RS485_TXE)
    RS485_TXE = x
  #endif
end macro

//
//------------------------------------------------------------------------
// helper functions
//------------------------------------------------------------------------
//
inline sub nop()
   asm
      nop
   end asm
end sub

//
//------------------------------------------------------------------------
// hardware setup/control functions
// **DEVICE DEPENDENT SECTION**
//------------------------------------------------------------------------
//

// init INTx for falling-edge trigger (for START bit detect)
sub init_intx()
    INTxIE = 0
    INTxIP = 1
  #if (HWSUART_INT_SELECT = 4) then
    IOCB = $10      // IOCB mask for IOCB4
  #endif
    INTxEDG = 0     // falling edge (this will also clear mismatch if using IOCB)
    nop()           // need one instruction for IOCB
    INTxIF = 0
end sub

//
// TMRx and CCPx peripheral init
//
// used to assign the TMR->CCP module selection
sub assign_ccp_tmr()
    select (CCP_NO)
        case 1      // CCP1 - CCPTMRS0 C1TSEL<1:0>:CCP1 Timer Selection bits
            CCPTMRS0 = CCPTMRS0 and not(%00000011)
            select (TMR_NO)
                case 1      // TMR1
                    CCPTMRS0 = CCPTMRS0 or %00000000
                case 3      // TMR3
                    CCPTMRS0 = CCPTMRS0 or %00000001
                case 5      // TMR5
                    CCPTMRS0 = CCPTMRS0 or %00000010
            end select
        case 2      // CCP2 - CCPTMRS0 C2TSEL<1:0>:CCP2 Timer Selection bits
            CCPTMRS0 = CCPTMRS0 and not(%00011000)
            select (TMR_NO)
                case 1      // TMR1
                    CCPTMRS0 = CCPTMRS0 or %00000000
                case 3      // TMR3
                    CCPTMRS0 = CCPTMRS0 or %00001000
                case 5      // TMR5
                    CCPTMRS0 = CCPTMRS0 or %00010000
            end select
    end select
end sub

sub init_tmr_ccp()
    TMRxIE = 0
    CCPxIE = 0

    // set timer 16-bit mode, internal counter (Tosc/4), 1:1 prescaler (off)
    TxCON = TMR_16BIT_INT_PRESCALER_OFF
    write_tmr16(0)

    // assign the TMR->CCP module selection
    assign_ccp_tmr()

    // set CCP compare mode, special event trigger, reset timer
    CCPxCON = CCP_COMPARE_SPEC_EVENT_TRIG
end sub

// setup CCPx module for 'wCount' period ('passed' in the CCPx registers)
// this is used to time the uart bit rate for RX and TX data bits,
// and the initial half-bit time for an rx START bit
inline sub setup_ccp(wCount as CCPx)
    CCPxIE = 0
    INTxIE = 0
    TMRxON = 0
    write_tmr16(0)
    CCPxIF = 0
end sub


// reload TMRx/CCPx count
// this routine is used after the rx START bit has been detected to setup
// the brg for a full bit time
// RELOAD_OVERHEAD lets you adjust the first cycle to account for the
// overhead in the routines
const RELOAD_OVERHEAD = 26      // number of instruction cycles
macro reload_ccp(wCount)
    TMRxON = 0
    TMRxH = 0
    TMRxL = RELOAD_OVERHEAD
    CCPRxH = wCount.bytes(1)
    CCPRxL = wCount.bytes(0)
    CCPxIF = 0
    TMRxON = 1
end macro

// setup to detect a START bit
sub enable_startbit_detect()
    setup_ccp(tmr_data.half_bit_time)   // preload CCP with half a bit time
    status_reg.rx = 0
    status_reg.wait_for_start = 1
    bit_count = 0
    rxd = 0
  #if (HWSUART_INT_SELECT = 4) then
    INTxEDG = 0     // clear mismatch if using IOCB
    nop()           // need one instruction for IOCB
  #endif
    INTxIF = 0
    INTxIE = 1
end sub

//
//------------------------------------------------------------------------
// public tx and rx routines
//------------------------------------------------------------------------
//

//
// check to see if there is data available in the rx buffer
// returns true if a character is in the buffer, false otherwise
//
public function data_available() as boolean
    result = rxhead <> rxtail
end function

//
// return the number of bytes in the receive buffer
//
public function buffer_count() as byte
    dim thead as byte

    thead = rxhead                          // read current head
    if (thead >= rxtail) then               // can just subtract
        result = thead - rxtail
    else                                    // must account for circular wrap
        result = RX_BUFFER_SIZE - rxtail    // num bytes from tail to end
        result = result + thead             // add in head
    endif
end function

// rx buffer full/empty checks
public function is_full() as boolean
    result = false
    if (buffer_count() = RX_BUFFER_SIZE-1) then
        result = true
    endif        
end function

public function is_empty() as boolean
    result = true
    if (data_available()) then
        result = false
    endif
end function

//
// check for input buffer overrun
// buffer overflow is sticky... once set you must reset the buffer to enable reception
// see flush_input() or reset()
//
public function overrun() as boolean
    result = status_reg.overrun
end function

public function frame_err() as boolean
    result = status_reg.frame_err
end function

//
// reset the uart
//
public sub reset()
    INTxIE = 0                  // disable the interrupts
    CCPxIE = 0

    rxhead = 0                  // init buffer
    rxtail = 0
    clear(rxbuf)

    status_reg.b = 0            // and status

    high(TXD_PIN)               // make sure txd is idle
    enable_startbit_detect()    // setup for start bit detect
end sub

//
// check to see if the uart is busy
//
inline function is_busy() as boolean
    result = (status_reg.rx = 1) or (status_reg.tx = 1)
end function

//
// flush the input buffer
//
public sub flush_input()
    // wait for uart to be available
    while (is_busy())
    end while

    reset()
end sub

//
// read a byte from the buffer
// NOTE: this function is non-blocking... data_available() MUST be called 
// (and return true) before calling this routine
//
public function read_byte() as byte
    dim tmptail as byte

    // calculate buffer index
    // tmptail = (rxtail + 1) and RX_BUFFER_MASK
    // done as separate statements so as to not require the use of a temp
    tmptail = rxtail
    inc(tmptail)
    tmptail = tmptail and RX_BUFFER_MASK

    // get buffer data
    result = rxbuf(tmptail)

    // update the new index
    rxtail = tmptail
end function

//
// write a byte to the uart
//
public sub write_byte(b as byte)
    // wait for uart to be available (ie not receiving)
    while (is_busy())
    end while

    // disable rx intr
    INTxIE = 0

    // set rs485 dir to tx (if used)
    RS485_TX_ENABLE(1)

    setup_ccp(tmr_data.bit_time)

    // setup tx data
    txd = b
    txd_out = 0         // setup START bit
    bit_count = 0

    // initiate transmission
    status_reg.wait_for_start = 0
    status_reg.rx = 0
    status_reg.tx_send_stop = 0
    status_reg.tx_done = 0
    status_reg.tx = 1

    // enable intr and start the baud rate timer
    CCPxIE = 1
    TMRxON = 1

    // wait for the byte to transmit...
    while (status_reg.tx_done = 0)
    end while

    // set rs485 dir to rx
    RS485_TX_ENABLE(0)

    // wait any additional character pacing delay
    delayus(tx_pacing)
end sub

// get bit or char time, in usecs
// can be used to set the tx_pacing
// ie to add an addtl stop bit: tx_pacing = bit_time()
public function bit_time() as word
    result = (1000000/tmr_data.baudrate)+1  // in usecs
end function

public function char_time() as word
    result = 10000000/tmr_data.baudrate     // in usecs
end function

//
//------------------------------------------------------------------------
// intr handler
//------------------------------------------------------------------------
//
// isr helper - add rxb byte to the rx buffer
// note: this routine is used by the receive isr, so 'inline' for speed.
// it is declared with the 'access' attribute to get local vars into
// the access bank (and a separate stack frame)
//
private inline sub access add_rxb_to_buffer()
    dim
        tmphead as byte,
        save_fsr0 as word

    // get next buffer index
    // tmphead = (rxhead + 1) and RX_BUFFER_MASK
    // done as separate statements so as to not require the use of a temp
    tmphead = rxhead
    inc(tmphead)
    tmphead = tmphead and RX_BUFFER_MASK

    // check to see if there's room in the input buffer
    if (tmphead = rxtail) then
        status_reg.overrun = true   // no room... byte is lost
    endif

    // if there's room in the buffer then add the char (rxb) 
    if (not status_reg.overrun) then
        save_fsr0 = FSR0            // save context...array access uses fsr0
        rxbuf(tmphead) = rxb        // add rxb into the buffer
        rxhead = tmphead            // and update the buffer head index
        FSR0 = save_fsr0            // restore context
    endif
end sub

//
// isr
//
interrupt access hwsuart_isr(PRIORITY_LEVEL)
    dim rxd_in as bit       // rx IO pin input sample

    // check for a rx start bit falling edge
    if (status_reg.wait_for_start = 1) then
        if (INTxIF = 1) then
            HWSUART_DEBUG_PIN_ON()
          #if (HWSUART_INT_SELECT = 4)
            if (INTxEDG = 1) then    // looking for a falling edge...
                nop()
                INTxIF = 0          // clear IF and exit
                exit
            endif
          #endif    // INT_SELECT = 4
            TMRxON = 1
            HWSUART_DEBUG_PIN_OFF()
            INTxIF = 0      // TIME: rx start bit falling edge to TMRxON = ~15 instruction cycles
            INTxIE = 0
            CCPxIE = 1     // ccp is already setup for halfbit time. just enable it
            status_reg.wait_for_start = 0
            status_reg.rx = 1
        endif
    endif

    // timing rx/tx data bits
    if ((CCPxIF = 1) and (CCPxIE = 1)) then
        CCPxIF = 0
        if (status_reg.rx = 1) then             // we're receiving...
            HWSUART_DEBUG_PIN_ON()              // TIME: from start bit TMRxON to data sample= ~13 instruction cycles
            rxd_in = RXD_PIN                    // sample RXD input pin
            if (bit_count = 0) then             // start bit handling
                reload_ccp(tmr_data.bit_time)       // set CCP back to full bit time
                if (rxd_in = 0) then                // start bit is ok
                    inc(bit_count)
                else                                // invalid start bit
                    status_reg.frame_err = true
                    enable_startbit_detect()
                endif
            elseif (bit_count = 9) then         // stop bit
                rxb = rxd                           // get a copy of rxd so we can setup for next byte
                enable_startbit_detect()            // setup for START while we buffer this byte
                if (rxd_in = 1) then                // stop bit ok
                    add_rxb_to_buffer()             // put data in buffer
                else                                // invalid stop bit
                    status_reg.frame_err = true
                endif
            else                                // data bits
                rxd = rxd >> 1                      // prepare for new bit
                if (rxd_in = 1) then                // add in new bit
                    rxd.bits(7) = 1
                endif
                inc(bit_count)
            endif
            HWSUART_DEBUG_PIN_OFF()
        elseif (status_reg.tx = 1) then         // we're transmitting...
            if (txd_out = 1) then                   // output previously setup pin data to TXD
                TXD_PIN = 1
            else
                TXD_PIN = 0
            endif
            inc(bit_count)
            if (status_reg.tx_send_stop = 1) then   // we just sent out stop bit
                status_reg.tx = 0
                status_reg.tx_done = 1
                enable_startbit_detect()            // setup to receive
            elseif (bit_count = 9) then
                txd_out = 1                         // send stop bit next time
                status_reg.tx_send_stop = 1
            else                                    // setup data bit for next time
                if (txd.bits(0) = 1) then
                    txd_out = 1
                else
                    txd_out = 0
                endif
                txd = txd >> 1
            endif
        endif
    endif
end interrupt

//
//------------------------------------------------------------------------
// high-level write functions (single byte and string)
//------------------------------------------------------------------------
//
sub WriteItem(pValue as WREG)
    write_byte(pValue)
end sub

sub WriteItem(pText as string)
    FSR0 = addressof(pText)
    while (INDF0 <> 0)
        write_byte(POSTINC0)
    end while
end sub

public compound sub Write(WriteItem)

//
//------------------------------------------------------------------------
// high-level read functions (single byte and string)
//------------------------------------------------------------------------
//
inline sub clrWDT()
    asm
        CLRWDT
    end asm
end sub

// common alias name for read string terminator
public dim
   ReadTerminator as rx_term
            
// read a byte from the UART (blocking)
public function ReadByte() as byte       
    repeat
        clrWDT
    until (data_available())
    result = read_byte()
end function

sub ReadItem(byref pValue as byte)
    pValue = ReadByte()
end sub

// read a string from the UART (blocking)
// use ReadTerminator to specify the input string terminator (default=null)
sub ReadItem(byref pText as string)
    dim TextPtr as POSTINC0
    dim TextAddr as FSR0
    dim Value as byte

    TextAddr = addressof(pText)
    Value = ReadByte()
    while (Value <> byte(ReadTerminator))
        TextPtr = Value
        Value = ReadByte()
    end while
    TextPtr = null
end sub

public compound sub Read(ReadItem)

public sub ClearOverrun()
    if (overrun()) then
        flush_input()
    endif
end sub    

// wait for a byte value to be received, with timeout value in milliseconds                                                   *
public function WaitForTimeout(pValue as byte, pTimeout as word) as boolean
    dim Counter as byte
    dim Timeout as word

    ClearOverrun()
    Result = false
    Counter = 10
    repeat
        Timeout = pTimeout
        while (Timeout > 0)
            if (data_available()) then
                if (read_byte() = pValue) then
                    Result = true
                    exit
                endif
            endif
            delayus(100)
            dec(Timeout)
        end while
        dec(Counter)
    until (Counter = 0)
end function

// wait for array of pCount bytes
// the incoming data is stored in pArray
public sub WaitForCount(byref pArray() as byte, pCount as word)
    dim Index as word

    ClearOverrun()
    Index = 0
    while (pCount > 0)
        pArray(Index) = ReadByte()
        dec(pCount)
        inc(Index)
    end while
end sub

// checks to see if a byte value has been received, with timeout in milliseconds
public function DataAvailableTimeout(pTimeout as word) as boolean
    dim Counter as byte
    dim Timeout as word

    ClearOverrun()
    Counter = 10
    repeat
        Timeout = pTimeout
        while (Timeout > 0) and not data_available()
            delayus(100)
            dec(Timeout)
        end while
        dec(Counter)
    until (Counter = 0) or data_available()
    result = data_available()
end function

//
//------------------------------------------------------------------------
// initialization functions
//------------------------------------------------------------------------
//

// used to check basic bit timing
// monitor TXD using a scope and measure low/high bit times
public sub check_bit_timing()
    // infinite loop sending 'U' ($55) char
    while (true)
      #if isoption(HWSUART_DEBUG_PIN)
        high(DEBUG_PIN)
      #endif
      write_byte($55)
      #if isoption(HWSUART_DEBUG_PIN)
        low(DEBUG_PIN)
      #endif
      delayus(100)
    end while
end sub

//
// setup io pins
//
inline sub init_io_pins()
    high(TXD_PIN)
    input(RXD_PIN)

  #if isoption(HWSUART_RS485_TXE)
    low(RS485_TXE)      // set rs485 to rx
    RS485_TX_ENABLE(0)
  #endif

  #if isoption(HWSUART_DEBUG_PIN)
    low(DEBUG_PIN)
  #endif
end sub

//
// main intialization function
// returns true=ok, false=invalid/unsupported baudrate
//
public function initialize(baudrate as longword = 9600) as boolean
    INTxIE = 0          // disable tmr intr
    CCPxIE = 0
    init_io_pins()         // setup io pins
    init_intx()
    init_tmr_ccp()

    // assume ok
    result = true

    // set timer data
    clear(tmr_data)
    select (baudrate)
        case 1200
            tmr_data.bit_time = BRG_1200
            tmr_data.half_bit_time = HBT_1200
        case 2400
            tmr_data.bit_time = BRG_2400
            tmr_data.half_bit_time = HBT_2400
        case 4800
            tmr_data.bit_time = BRG_4800
            tmr_data.half_bit_time = HBT_4800
        case 9600
            tmr_data.bit_time = BRG_9600
            tmr_data.half_bit_time = HBT_9600
        case 19200
            tmr_data.bit_time = BRG_19200
            tmr_data.half_bit_time = HBT_19200
        case 38400
            tmr_data.bit_time = BRG_38400
            tmr_data.half_bit_time = HBT_38400
        case 57600
            tmr_data.bit_time = BRG_57600
            tmr_data.half_bit_time = HBT_57600
        case 115200
            tmr_data.bit_time = BRG_115200
            tmr_data.half_bit_time = HBT_115200
      #if (_clock = 64)
        case 230400
            tmr_data.bit_time = BRG_230400
            tmr_data.half_bit_time = BRG_230400/2
      #endif
        else        // invalid baudrate
            tmr_data.half_bit_time = 0
    end select
    tmr_data.baudrate = baudrate

    // check resulting half-bit time
    if (tmr_data.half_bit_time = 0) then            // unsupported
        result = false
    elseif (tmr_data.half_bit_time = 1) then        // user override
        result = true                               // generate "immediate" start bit
    elseif (tmr_data.half_bit_time <= SBOH) then    // this likely won't work
        result = false                          
        tmr_data.half_bit_time = 1
    else                                            // ok... correct for overhead
        tmr_data.half_bit_time = tmr_data.half_bit_time - SBOH
    endif        

    // set intr priority (ipHigh is recommended)
    if (PRIORITY_LEVEL = 2) then
        INTxIP = 1
        CCPxIP = 1
    else
        INTxIP = 0
        CCPxIP = 0
    endif

    // reset suart state... enables startbit detect IE
    reset()

    // if valid then enable intr
    if (result) then
        enable(hwsuart_isr)
    endif
end function

//
//------------------------------------------------------------------------
// module initialization
//------------------------------------------------------------------------
//
init_io_pins()
rx_term = null
tx_pacing = DEFAULT_TX_PACING   // default pacing delay, in usecs

end module
