//
//------------------------------------------------------------------------------
//  Name    : hwsuart_T0CKI.bas
//  Author  : Jerry Messina
//  Date    : 10/26/2018
//  Version : 1.2.0
//------------------------------------------------------------------------------
//
//  Hardware-assisted software uart (hwsuart using tmr0 T0CKI)
//
//  This module is an interrupt-driven 'software' uart that uses hardware peripheral
//  functions to assist in the generation of the serial data timing and detection of
//  the RX start bit. In addition, it provides an input RX buffer to hold incoming data
//  allowing the module to be used for asyncronous reception (which is something most
//  software uarts cannot do).
//
//  The module uses TMR0 in both internal timer and external counter modes for the
//  receive start bit edge detect and bit time generation for both RX and TX data.
//  The module relies on TMR0 interrupts to provide the bit timing, and should be
//  assigned to the high-priority intr to maintain timing.
//
//  It requires that the RX data in signal be connected to T0CKI (typically RA4).
//  This is required as TMR0 is used in external falling edge trigger mode to detect 
//  the beginning of a START bit. The TX data out can reside on any port pin.
//
//  The hwsuart operates in half-duplex mode only (you would need a separate rx and 
//  tx bit timer to support full-duplex operation). For reception, it provides an
//  interrupt-driven 4 to 128 char circular buffer for received chars, controllable 
//  by the '#option HWSUART_RX_BUFFER_SIZE' setting. 
//
//  Be aware that since it operates in half-duplex mode, it's possible to drop incoming
//  characters or even get a false corrupt byte (a partial char) if the module is busy 
//  transmitting when a char comes in.
//
//------------------------------------------------------------------------------
//
module hwsuart

// input buffer size (must be a power of 2 between 4 and 128)
#option HWSUART_RX_BUFFER_SIZE = 8      // set default buffer size
#if not (HWSUART_RX_BUFFER_SIZE in (4, 8, 16, 32, 64, 128))
  #error HWSUART_RX_BUFFER_SIZE, "invalid buffer size option"
#endif
public const RX_BUFFER_SIZE = HWSUART_RX_BUFFER_SIZE

//
// hardware IO pin definitions
// note: if using an analog/digital pin be sure to set the pins to digital mode, 
// and disable any slew-rate control (if applicable for your device)
//

// TXD and RXD IO pin definitions
// software uart TXD can be any pin
#option HWSUART_TXD_PIN = PORTC.6
#if isvalidportpin(HWSUART_TXD_PIN)
  dim TXD_PIN as HWSUART_TXD_PIN.HWSUART_TXD_PIN@
#else
  #error HWSUART_TXD_PIN, "invalid port pin option"
#endif

// software uart RXD must be the T0CKI pin
// typically T0CKI is PORTA.4 (check the datasheet for your device)
// if using PPS, you can specify a different pin but you must map the 
// T0CKI function to the pin before calling initialize()
#option HWSUART_RXD_PIN = PORTA.4
#if isvalidportpin(HWSUART_RXD_PIN)
  dim RXD_PIN as HWSUART_RXD_PIN.HWSUART_RXD_PIN@
#else
  #error HWSUART_RXD_PIN, "invalid port pin option"
#endif

// enable debug pin output (optional)
// the debug pin can be used to measure the rx bit timing. it is pulsed when
// the rx startbit is detect and then asserted high when the rx pin is sampled
// and remain high for the duration of the rx routine. 
// to measure the rx timing, send alternating bit pattern (ie 'U' $55) and look
// at the DEBUG pin vs the RXD pin on a scope. ideally the RXD pin will be sampled
// at 50%-75% of the bit time.
#if isoption(HWSUART_DEBUG_PIN)
  #if isvalidportpin(HWSUART_DEBUG_PIN)
    #warning "enabling HWSUART_DEBUG_PIN output"
    dim DEBUG_PIN as HWSUART_DEBUG_PIN.HWSUART_DEBUG_PIN@    // debug output (for scope)
  #else
    #error HWSUART_DEBUG_PIN, "invalid port pin option"
  #endif
#endif  // isoption(HWSUART_DEBUG_PIN)

// TX pacing default, in usecs
// here you can add additional time after the tx stop bit is sent in case the receiving
// device needs some time in between bytes. can be changed at runtime by setting 'tx_pacing'
#option HWSUART_TX_PACING = 0
const DEFAULT_TX_PACING as word = HWSUART_TX_PACING

// interrupt priority level
// since the bit timing is critical, only the high-priority intr should be used
#option HWSUART_PRIORITY_LEVEL = 2      // 1=low, 2=high
#if not(HWSUART_PRIORITY_LEVEL in (1, 2))
  #error HWSUART_PRIORITY_LEVEL, "intr priority must be 1 or 2"
#endif  
const PRIORITY_LEVEL = HWSUART_PRIORITY_LEVEL

// TMR0 control register format (device-dependent)
// K40 and Q10 use different register settings than most all other devices...
#if _device in(18F24K40, 18F25K40, 18F26K40, 18F27K40) or _
    _device in(18F45K40, 18F46K40, 18F47K40) or _
    _device in(18F65K40, 18F66K40, 18F67K40) or _
    _device in(18F24Q10, 18F25Q10, 18F26Q10, 18F27Q10) or _
    _device in(18F45Q10, 18F46Q10, 18F47Q10)
  #option HWSUART_TMR0_FORMAT = 2       // dual T0CON0, T0CON1
#else
  #option HWSUART_TMR0_FORMAT = 1       // single T0CON (most devices)
#endif
#if not(HWSUART_TMR0_FORMAT in (1, 2))
  #error HWSUART_TMR0_FORMAT, "tmr0 format must be 1 or 2"
#endif  

//
//------------------------------------------------------------------------
// uart state/status register
//------------------------------------------------------------------------
//
public structure suart_status_reg_t
    b as byte
    wait_for_start  as b.bits(0)        // edge detect enabled (idle, waiting)
    rx              as b.bits(1)        // receiving (busy)
    tx              as b.bits(2)        // tx request (busy)
    tx_send_stop    as b.bits(3)        // tx sending stop bit
    tx_done         as b.bits(4)        // tx complete
    spare1          as b.bits(5)        // unused
    frame_err       as b.booleans(6)    // invalid start/stop bit
    overrun         as b.booleans(7)    // too many rcv chars in rxbuf. input blocks
end structure

public dim status_reg as suart_status_reg_t

//
//------------------------------------------------------------------------
// tx/rx variables and buffers
//------------------------------------------------------------------------
//
// the receive data buffer is implemented as a simple circular queue.
// data is placed into the queue at 'rxhead', and is read from the queue at 'rxtail'.
// in this simple implementation, no additional status tracking flags are used (ie
// no buffer full/empty), and as a result you lose one position in the buffer which
// is used to signify that the buffer is empty (when rxhead = rxtail). The size of the
// buffer can be changed with the '#option HWSUART_RX_BUFFER_SIZE'. The buffer size
// defaults to 8, so the buffer can hold up to 7 bytes of data.
//
public const HWSUART_MAX_CHARS = RX_BUFFER_SIZE-1   // max number of chars in the buffer
const RX_BUFFER_MASK = RX_BUFFER_SIZE-1             // buffer size MUST be a power of 2

dim
    rxd as byte,                    // rx in shift reg
    txd as byte,                    // tx out shift reg
    bit_count as byte,              // current bit number (start + 8 data + stop)
    txd_out as bit                  // tx IO pin output data

// baud rate timer settings
structure timer_data_t
    baudrate            as longword // current baudrate
    bit_time            as word     // full bit time in tmr0 ticks
    half_bit_time       as word     // initial half-bit time in tmr0 ticks
end structure

dim
    tmr_data as timer_data_t        // tmr0 count values (from table)

// additional time after the tx stop bit is sent
public dim
    tx_pacing as word               // delay, in usecs
    
// rx buffer
dim
    rxb as byte,                            // rx byte to add to buffer
    rxhead as byte,                         // buf indexes... data goes in at the head
    rxtail as byte,                         // and is read out the tail
    rxbuf(RX_BUFFER_SIZE) as byte           // receive buffer

public dim
   rx_term as char                          // read string terminator

//
//------------------------------------------------------------------------
// timer setting (device dependent)
//------------------------------------------------------------------------
//
#if (HWSUART_TMR0_FORMAT = 1)       // single T0CON (default)
// these are typ settings for most devices... check they match your device

// T0CON settings
// external edge mode setup
//    bit 7: T0CON_TMR0ON = 0    ' stop timer to change mode
//    bit 6: T0CON_T08BIT = 0    ' 16-bit
//    bit 5: T0CON_T0CS = 1      ' external
//    bit 4: T0CON_T0SE = 1      ' falling edge (high to low)
//    bit 3: T0CON_PSA = 1       ' prescaler off (1:1)
//    b 2-0: prescaler

// internal timer mode setup
//    bit 7: T0CON_TMR0ON = 0    ' stop timer to change mode
//    bit 6: T0CON_T08BIT = 0    ' 16-bit
//    bit 5: T0CON_T0CS = 0      ' internal instruction cycle
//    bit 4: T0CON_T0SE = 0      ' don't care
//    bit 3: T0CON_PSA = 1       ' 1=prescaler off (1:1)
//    b 2-0: T0CON_PS2-PS0

// T0CON register settings
dim T0CON_REG as T0CON
const
    TMR0_FALLING_EDGE_MODE as byte = %00111000,     // ext falling edge
    TMR0_TIMER_MODE        as byte = %00001000      // internal instruction cycle clock

// TMR0 on/off bit
dim TMR0ON as T0CON.bits(7)

// peripheral interrupt bits for TMR0
dim
    TMR0IF as INTCON.bits(2),
    TMR0IE as INTCON.bits(5),
    TMR0IP as INTCON2.bits(2)

#elseif (HWSUART_TMR0_FORMAT = 2)       // dual T0CON0, T0CON1
// K40, Q10 format
// T0CON0
//    bit 7: T0EN
//    bit 6: 
//    bit 5: 
//    bit 4: T016BIT = 1
//    bits 3:0: T0OUTPS = 000 (1:1 postscaler)

// T0CON1
//    bit 7:5:  T0CS[2:0]    clock source select 
//                           010=Fosc/4 (INT), 001=pin selected by T0CKIPPS (inverted)
//    bit 4:    T0ASYNC = 0
//    bits 3:0: T0CKPS = 000 (1:1 prescaler)

dim T0CON_REG as T0CON1
const
    TMR0_FALLING_EDGE_MODE as byte = %00110000,     // ext falling edge, async
    TMR0_TIMER_MODE        as byte = %01000000      // internal instruction cycle clock

// T0CON0 setup
const
    T0CON0_MODE as byte = %00010000                 // 16-bit mode, postscale 1:1
    
// TMR0 on/off bit
dim TMR0ON as T0CON0.bits(7)

// peripheral interrupt bits for TMR0
dim
    TMR0IF as PIR0.bits(5),
    TMR0IE as PIE0.bits(5),
    TMR0IP as IPR0.bits(5)
#endif  // HWSUART_TMR0_FORMAT

//
//------------------------------------------------------------------------
// uart baudrate bit times
//------------------------------------------------------------------------
//
// uart baud rate bit times (in usecs)
//   9600 = 104.16us/bit
//  19200 =  52.08us
//  28800 =  34.72us
//  38400 =  26.04us
//  57600 =  17.36us
// 115200 =   8.68us

// clock dependent baud rate timings
const CPU_CLK = _clock                      // system clock freq
const TMR_CLK = (CPU_CLK * 1000000.0)/4     // timer clock, in MHz (FOSC/4)

// uart baud rate bit times, in tmr clocks (ie instruction cycles)
// bit_time = (_clock/4)/baudrate
const
   BRG_1200     as word = TMR_CLK / 1200,           // 1200 baud
   BRG_2400     as word = TMR_CLK / 2400,           // 2400 baud
   BRG_4800     as word = TMR_CLK / 4800,           // 4800 baud
   BRG_9600     as word = TMR_CLK / 9600,           // 9600 baud
   BRG_19200    as word = TMR_CLK / 19200,          // 19200 baud
   BRG_38400    as word = TMR_CLK / 38400,          // 38400 baud
   BRG_57600    as word = TMR_CLK / 57600,          // 57600 baud (min clock 16MHz)
   BRG_115200   as word = TMR_CLK / 115200          // 115200 baud (min clock 32MHz)

// clock dependent baud rate timings
// these values are directly in instruction cycles, and will be subtracted from 65536 (since 
//  the counter counts up and generates an intr on rollover to 0) when the data structure is
//  loaded during setup
// they should be adjusted to take into account the overhead in the isr, which is done
//  when the 'tmr_data' structure is loaded in initialize(). Some combinations of slower
//  clock speeds and faster baud rates require skipping the timing of the half-bit time 
//  altogether, and are set to 0
// the easiest way to determine/verify these values is to enable the DEBUG_PIN output
//  using '#option HWSUART_DEBUG_ENABLE = true' and look at the timing of RXD vs the DEBUG_PIN
//
const 
    ISR_OVERHEAD = 16,                      // number of clock cycles of isr overhead
    ISR_HB_OVERHEAD = ISR_OVERHEAD + 24     // number of clock cycles of isr overhead for half-bit
    
// these are supported at all clock frequencies with no special considerations
const
   HBT_1200  as word = BRG_1200/2,
   HBT_2400  as word = BRG_2400/2,
   HBT_4800  as word = BRG_4800/2,
   HBT_9600  as word = BRG_9600/2,
   HBT_19200 as word = BRG_19200/2,
   HBT_38400 as word = BRG_38400/2

// bit_time and half_bit_time data (timer_data_t), in clock cycles
#if (_clock = 8)
const
   HBT_57600 as word = 0,               // not supported
   HBT_115200 as word = 0               // not supported

#elseif (_clock = 16)
const
   HBT_57600 as word = 1,               // supported, but skip half bit time
   HBT_115200 as word = 0               // not supported

#elseif (_clock = 32)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = BRG_115200/2

#elseif (_clock = 64)
const
   HBT_57600 as word = BRG_57600/2,
   HBT_115200 as word = BRG_115200/2

#else
  #error "unsupported HWSUART clock freq"
#endif

//
//------------------------------------------------------------------------
// debug pin support
//------------------------------------------------------------------------
//
inline sub HWSUART_DEBUG_PIN_ON()
  #if isoption(HWSUART_DEBUG_PIN)
    DEBUG_PIN = 1
  #endif
end sub

inline sub HWSUART_DEBUG_PIN_OFF()
  #if isoption(HWSUART_DEBUG_PIN)
    DEBUG_PIN = 0
  #endif
end sub

//
//------------------------------------------------------------------------
// TMR0 routines
//------------------------------------------------------------------------
// initially the timer is set to count 1 external falling edge (start bit)
// once the start bit is detected we switch it to internal timer mode
// to provide the baudrate bit timer 
// most of these are 'inline' or macros for speed
//

// write tmr0 as a 16-bit timer (high byte, then low byte)
public macro write_tmr0(wval)
    TMR0H = (wval >> 8)
    TMR0L = byte(wval)
end macro    

// set TMR0 to count 1 external falling edge (start bit)
sub enable_startbit_detect()
    T0CON_REG = TMR0_FALLING_EDGE_MODE
    write_tmr0($ffff)       // count 1 edge
    TMR0IF = 0              // reset intr flag

    status_reg.rx = 0
    status_reg.wait_for_start = 1
    bit_count = 0
    rxd = 0

    TMR0ON = 1              // turn on timer
end sub

// set tmr for half of a bit time (switch to timer mode) 
inline sub tmr0_half_bit()
    T0CON_REG = TMR0_TIMER_MODE
    write_tmr0(tmr_data.half_bit_time)
    TMR0IF = 0          // reset intr flag
    TMR0ON = 1          // turn on timer
end sub

// set tmr for full bit time (switch to timer mode) 
inline sub tmr0_full_bit()
    T0CON_REG = TMR0_TIMER_MODE
    write_tmr0(tmr_data.bit_time)
    TMR0IF = 0          // reset intr flag
    TMR0ON = 1          // turn on timer
end sub

//
//------------------------------------------------------------------------
// public tx and rx routines
//------------------------------------------------------------------------
//

//
// check to see if there is data available in the rx buffer
// returns true if a character is in the buffer, false otherwise
//
public function data_available() as boolean
    result = rxhead <> rxtail
end function

//
// return the number of bytes in the receive buffer
//
public function buffer_count() as byte
    dim thead as byte

    thead = rxhead                          // read current head
    if (thead >= rxtail) then               // can just subtract
        result = thead - rxtail
    else                                    // must account for circular wrap
        result = RX_BUFFER_SIZE - rxtail    // num bytes from tail to end
        result = result + thead             // add in head
    endif
end function

// rx buffer full/empty checks
public function is_full() as boolean
    result = false
    if (buffer_count() = RX_BUFFER_SIZE-1) then
        result = true
    endif        
end function

public function is_empty() as boolean
    result = true
    if (data_available()) then
        result = false
    endif
end function

//
// check for input buffer overrun
// buffer overflow is sticky... once set you must call flush_input()
// to enable reception again
//
public function overrun() as boolean
    result = status_reg.overrun
end function

public function frame_err() as boolean
    result = status_reg.frame_err
end function

//
// reset the uart
//
public sub reset()
    dim intena as bit

    intena = TMR0IE         // get current IE
    TMR0IE = 0              // disable tmr intr

    rxhead = 0              // init buffer
    rxtail = 0
    clear(rxbuf)

    status_reg.b = 0        // and status

    high(TXD_PIN)               // make sure txd is idle
    enable_startbit_detect()    // setup for start bit detect

    if (intena = 1) then    // restore intr enable setting
        TMR0IE = 1
    endif
end sub

//
// check to see if the uart is busy
//
inline function is_busy() as boolean
    result = (status_reg.rx = 1) or (status_reg.tx = 1)
end function

//
// flush the input buffer
//
public sub flush_input()
    // wait for uart to be available
    while (is_busy())
    end while

    reset()
end sub

//
// read a byte from the buffer
// NOTE: this function is non-blocking... data_available() MUST be called 
// (and return true) before calling this routine
//
public function read_byte() as byte
    dim tmptail as byte

    // calculate buffer index
    // tmptail = (rxtail + 1) and RX_BUFFER_MASK
    // done as separate statements so as to not require the use of a temp
    tmptail = rxtail
    inc(tmptail)
    tmptail = tmptail and RX_BUFFER_MASK

    // get buffer data
    result = rxbuf(tmptail)

    // update the new index
    rxtail = tmptail
end function

//
// write byte
//
public sub write_byte(b as byte)
    // wait for uart to be available (ie not receiving)
    while (is_busy())
    end while

    // disable tmr intr while we set things up
    TMR0IE = 0          // disable tmr intr
    TMR0ON = 0          // stop timer

    // setup tx data
    txd = b
    txd_out = 0         // setup START bit
    bit_count = 0

    // initiate transmission
    status_reg.wait_for_start = 0
    status_reg.rx = 0
    status_reg.tx_send_stop = 0
    status_reg.tx_done = 0
    status_reg.tx = 1

    // manually generate a timer intr to get the ball rolling
    // the isr will send out the start bit, data bits, and the stop bit
    TMR0IF = 1              // trigger it
    TMR0IE = 1              // and enable tmr intr

    // wait for the byte to transmit...
    while (status_reg.tx_done = 0)
    end while

    // wait any additional character pacing delay
    delayus(tx_pacing)
end sub

// get bit or char time, in usecs
// can be used to set the tx_pacing
// ie to add an addtl stop bit: tx_pacing = bit_time()
public function bit_time() as word
    result = (1000000/tmr_data.baudrate)+1  // in usecs
end function

public function char_time() as word
    result = 10000000/tmr_data.baudrate     // in usecs
end function

//
//------------------------------------------------------------------------
// intr handler
//------------------------------------------------------------------------
//
// isr helper - add rxb byte to the rx buffer
// note: this routine is used by the receive isr, so 'inline' for speed.
// it is declared with the 'access' attribute to get local vars into
// the access bank (and a separate stack frame)
//
private inline sub access add_rxb_to_buffer()
    dim
        tmphead as byte,
        save_fsr0 as word

    // get next buffer index
    // tmphead = (rxhead + 1) and RX_BUFFER_MASK
    // done as separate statements so as to not require the use of a temp
    tmphead = rxhead
    inc(tmphead)
    tmphead = tmphead and RX_BUFFER_MASK

    // check to see if there's room in the input buffer
    if (tmphead = rxtail) then
        status_reg.overrun = true   // no room... byte is lost
    endif

    // if there's room in the buffer then add the char (rxb) 
    if (not status_reg.overrun) then
        save_fsr0 = FSR0            // save context...array access uses fsr0
        rxbuf(tmphead) = rxb        // add rxb into the buffer
        rxhead = tmphead            // and update the buffer head index
        FSR0 = save_fsr0            // restore context
    endif
end sub

//
// isr
//
interrupt access hwsuart_isr(PRIORITY_LEVEL)
    dim rxd_in as bit       // rx IO pin input sample

    // restart timer for a full bit period (we'll change it later if req'd)
    tmr0_full_bit()

    if (status_reg.wait_for_start = 1) then         // we've detected a start bit...
        // prepare to receive a byte
        HWSUART_DEBUG_PIN_ON()
        status_reg.wait_for_start = 0
        status_reg.rx = 1
        HWSUART_DEBUG_PIN_OFF()
        // setup the timer to put us in the middle of the start bit
        // we do this last so that we can account for the overhead of
        // the previous code to setup reception
        if (tmr_data.half_bit_time <> 0) then
            tmr0_half_bit()
            exit
        endif
    endif
    
    if (status_reg.rx = 1) then         // we're receiving...
        HWSUART_DEBUG_PIN_ON()
        rxd_in = RXD_PIN                    // read the RX data in pin
        if (bit_count = 0) then             // start bit handling
            if (rxd_in = 0) then                // start bit is ok
                inc(bit_count)
            else                                // invalid start bit
                status_reg.frame_err = true
                enable_startbit_detect()
            endif
        elseif (bit_count = 9) then         // stop bit
            rxb = rxd                           // get a copy of rxd so we can setup for next byte
            enable_startbit_detect()            // setup for START while we buffer this byte
            if (rxd_in = 1) then                // stop bit ok
                add_rxb_to_buffer()             // put data in buffer
            else                                // invalid stop bit
                status_reg.frame_err = true
            endif
        else                                // data bits
            rxd = rxd >> 1                      // prepare for new bit
            if (rxd_in = 1) then                // add in new bit
                rxd.bits(7) = 1
            endif
            inc(bit_count)
        endif
        HWSUART_DEBUG_PIN_OFF()
    elseif (status_reg.tx = 1) then         // we're transmitting...
        if (txd_out = 1) then                   // output previously setup pin data to TXD
            TXD_PIN = 1
        else
            TXD_PIN = 0
        endif
        inc(bit_count)
        if (status_reg.tx_send_stop = 1) then   // we just sent out stop bit
            status_reg.tx = 0
            status_reg.tx_done = 1
            enable_startbit_detect()            // setup to receive
        elseif (bit_count = 9) then
            txd_out = 1                         // send stop bit next time
            status_reg.tx_send_stop = 1
        else                                    // setup data bit for next time
            if (txd.bits(0) = 1) then
                txd_out = 1
            else
                txd_out = 0
            endif
            txd = txd >> 1
        endif
    endif
end interrupt

//
//------------------------------------------------------------------------
// high-level write functions (single byte and string)
//------------------------------------------------------------------------
//
sub WriteItem(pValue as WREG)
    write_byte(pValue)
end sub

sub WriteItem(pText as string)
    FSR0 = addressof(pText)
    while (INDF0 <> 0)
        write_byte(POSTINC0)
    end while
end sub

public compound sub Write(WriteItem)

//
//------------------------------------------------------------------------
// high-level read functions (single byte and string)
//------------------------------------------------------------------------
//
inline sub clrWDT()
    asm
        CLRWDT
    end asm
end sub

// common alias name for read string terminator
public dim
   ReadTerminator as rx_term
            
// read a byte from the UART (blocking)
public function ReadByte() as byte       
    repeat
        clrWDT
    until (data_available())
    result = read_byte()
end function

sub ReadItem(byref pValue as byte)
    pValue = ReadByte()
end sub

// read a string from the UART (blocking)
// use ReadTerminator to specify the input string terminator (default=null)
sub ReadItem(byref pText as string)
    dim TextPtr as POSTINC0
    dim TextAddr as FSR0
    dim Value as byte

    TextAddr = addressof(pText)
    Value = ReadByte()
    while (Value <> byte(ReadTerminator))
        TextPtr = Value
        Value = ReadByte()
    end while
    TextPtr = null
end sub

public compound sub Read(ReadItem)

public sub ClearOverrun()
    if (overrun()) then
        flush_input()
    endif
end sub    

// wait for a byte value to be received, with timeout value in milliseconds                                                   *
public function WaitForTimeout(pValue as byte, pTimeout as word) as boolean
    dim Counter as byte
    dim Timeout as word

    ClearOverrun()
    Result = false
    Counter = 10
    repeat
        Timeout = pTimeout
        while (Timeout > 0)
            if (data_available()) then
                if (read_byte() = pValue) then
                    Result = true
                    exit
                endif
            endif
            delayus(100)
            dec(Timeout)
        end while
        dec(Counter)
    until (Counter = 0)
end function

// wait for array of pCount bytes
// the incoming data is stored in pArray
public sub WaitForCount(byref pArray() as byte, pCount as word)
    dim Index as word

    ClearOverrun()
    Index = 0
    while (pCount > 0)
        pArray(Index) = ReadByte()
        dec(pCount)
        inc(Index)
    end while
end sub

// checks to see if a byte value has been received, with timeout in milliseconds
public function DataAvailableTimeout(pTimeout as word) as boolean
    dim Counter as byte
    dim Timeout as word

    ClearOverrun()
    Counter = 10
    repeat
        Timeout = pTimeout
        while (Timeout > 0) and not data_available()
            delayus(100)
            dec(Timeout)
        end while
        dec(Counter)
    until (Counter = 0) or data_available()
    result = data_available()
end function

//
//------------------------------------------------------------------------
// initialization functions
//------------------------------------------------------------------------
//

// used to check basic bit timing
// monitor TXD using a scope and measure low/high bit times
public sub check_bit_timing()
    // infinite loop sending 'U' ($55) char
    while (true)
      #if isoption(HWSUART_DEBUG_PIN)
        high(DEBUG_PIN)
      #endif
      write_byte($55)
      #if isoption(HWSUART_DEBUG_PIN)
        low(DEBUG_PIN)
      #endif
      delayus(100)
    end while
end sub

//
// setup io pins
//
inline sub init_io_pins()
    high(TXD_PIN)
    input(RXD_PIN)

  #if isoption(HWSUART_DEBUG_PIN)
    low(DEBUG_PIN)
  #endif
end sub

//
// main intialization function
// returns true=ok, false=invalid/unsupported baudrate
//
public function initialize(baudrate as longword = 9600) as boolean
    TMR0IE = 0          // disable tmr intr
    TMR0ON = 0          // and turn off timer
    
    init_io_pins()      // setup io pins

  #if (HWSUART_TMR0_FORMAT = 2)         // K40, Q10
    T0CON0 = T0CON0_MODE
  #endif
  
    // assume ok
    result = true

    // set timer data
    clear(tmr_data)
    select (baudrate)
        case 1200
            tmr_data.bit_time = BRG_1200
            tmr_data.half_bit_time = HBT_1200
        case 2400
            tmr_data.bit_time = BRG_2400
            tmr_data.half_bit_time = HBT_2400
        case 4800
            tmr_data.bit_time = BRG_4800
            tmr_data.half_bit_time = HBT_4800
        case 9600
            tmr_data.bit_time = BRG_9600
            tmr_data.half_bit_time = HBT_9600
        case 19200
            tmr_data.bit_time = BRG_19200
            tmr_data.half_bit_time = HBT_19200
        case 38400
            tmr_data.bit_time = BRG_38400
            tmr_data.half_bit_time = HBT_38400
        case 57600
            tmr_data.bit_time = BRG_57600
            tmr_data.half_bit_time = HBT_57600
        case 115200
            tmr_data.bit_time = BRG_115200
            tmr_data.half_bit_time = HBT_115200
        else        // unsupported baudrate
            result = false
    end select
    tmr_data.baudrate = baudrate
    
    // timer values must be subtracted from 65536 since the counter counts up
    // and generates an intr on rollover to 0
    if (tmr_data.bit_time <> 0) and (tmr_data.bit_time > ISR_OVERHEAD) then
        tmr_data.bit_time = 65536 - (tmr_data.bit_time - ISR_OVERHEAD)
    else
        result = false      // unsupported
    endif
    if (tmr_data.half_bit_time <> 0) then
        if (tmr_data.half_bit_time <= ISR_HB_OVERHEAD) then
            tmr_data.half_bit_time = 0
        else
            tmr_data.half_bit_time = 65536 - (tmr_data.half_bit_time - ISR_HB_OVERHEAD)
        endif
    else
        result = false      // unsupported
    endif

    // set intr priority (ipHigh is recommended)
    if (PRIORITY_LEVEL = 2) then
        TMR0IP = 1
    else
        TMR0IP = 0
    endif
    
    // reset suart state
    reset()
  
    // if valid then enable intr
    if (result) then
        TMR0IE = 1
        enable(hwsuart_isr)
    endif
end function

//
//------------------------------------------------------------------------
// module initialization
//------------------------------------------------------------------------
//
init_io_pins()
rx_term = null
tx_pacing = DEFAULT_TX_PACING   // default pacing delay, in usecs

end module
